@@ -29,14 +29,15 @@ use GNUpod::ArtworkDB;
 use Getopt::Long;
 use File::Copy;
 use File::Glob ':glob';
+use Date::Parse;
 
 use constant MEDIATYPE_PODCAST_AUDIO => 4;
 use constant MEDIATYPE_PODCAST_VIDEO => 6;
 
 use constant MACTIME => GNUpod::FooBar::MACTIME;
-use vars qw(%opts %dupdb_normal %dupdb_lazy %dupdb_podcast $int_count %podcast_infos %per_file_info);
+use vars qw(%opts %dupdb_normal %dupdb_lazy %dupdb_podcast $int_count %podcast_infos %podcast_channel_infos %per_file_info);
 
-print "gnupod_addsong.pl Version 0.99.6 (C) Adrian Ulrich\n";
+print "gnupod_addsong.pl Version 0.99.7 (C) Adrian Ulrich\n";
 
 $int_count = 3; #The user has to send INT (Ctrl+C) x times until we stop
 
@@ -46,11 +47,13 @@ GetOptions(\%opts, "version", "help|h", "mount|m=s", "decode|x=s", "restore|r",
                    "set-title|t=s", "set-artist|a=s", "set-album|l=s", "set-genre|g=s", "set-rating=i", "set-playcount=i",
                    "set-bookmarkable|b", "set-shuffleskip", "artwork=s",
                    "set-songnum", "playlist|p=s@", "reencode|e=i",
-                   "min-vol-adj=i", "max-vol-adj=i", "playlist-is-podcast", "set-compilation");
+                   "min-vol-adj=i", "max-vol-adj=i", "playlist-is-podcast", "podcast-files-limit=i", "podcast-cache-dir=s",
+                   "podcast-artwork", "set-compilation");
 
 GNUpod::FooBar::GetConfig(\%opts, {'decode'=>'s', mount=>'s', duplicate=>'b', model=>'s',
                                    'disable-v1'=>'b', 'disable-v2'=>'b', 'set-songnum'=>'b',
-                                   'min-vol-adj'=>'i', 'max-vol-adj'=>'i', 'automktunes'=>'b' },
+                                   'min-vol-adj'=>'i', 'max-vol-adj'=>'i', 'automktunes'=>'b', 
+                                   'podcast-files-limit'=>'i', 'podcast-cache-dir'=>'s', 'podcast-artwork'=>'b' },
                                    "gnupod_addsong");
 
 
@@ -66,7 +69,7 @@ if($opts{restore}) {
 	print "If you use --restore, you'll *lose* your playlists and cover artwork!\n";
 	print " Hit ENTER to continue or CTRL+C to abort\n\n";
 	<STDIN>;
-	@XFILES = bsd_glob("$opts{mount}/i*/Music/F*/*", GLOB_NOSORT)
+	@XFILES = bsd_glob("$opts{mount}/i*/Music/[Ff]*/*", GLOB_NOSORT)
 }
 elsif($ARGV[0] eq "-" && @ARGV == 1) {
 	print STDERR "Reading from STDIN, hit CTRL+D (EOF) when finished\n";
@@ -83,8 +86,8 @@ else {
 my $connection = GNUpod::FooBar::connect(\%opts);
 usage($connection->{status}."\n") if $connection->{status} || !@XFILES;
 
-my $AWDB  = GNUpod::ArtworkDB->new(Connection=>$connection, DropUnseen=>1);
-
+my $AWDB  = GNUpod::ArtworkDB->new(Connection=>$connection, DropUnseen=>($opts{'podcast-artwork'}?0:1));
+my $awdb_image_prepared = 0 ;
 
 my $exit_code = startup($connection,@XFILES);
 exit($exit_code);
@@ -111,10 +114,7 @@ sub startup {
 	}
 	else {
 		if($opts{artwork}) {
-			if( $AWDB->PrepareImage(File=>$opts{artwork}, Model=>$opts{model}) ) {
-				$AWDB->LoadArtworkDb or die "Failed to load artwork database\n";
-			}
-			else {
+			if ( ! add_image_to_awdb($opts{artwork})) {
 				warn "$0: Could not load $opts{artwork}, skipping artwork\n";
 				delete($opts{artwork});
 			}
@@ -186,7 +186,7 @@ sub startup {
 		$fh->{playcount}    = $opts{'set-playcount'}   if $opts{'set-playcount'};
 		$fh->{title}        = $opts{'set-title'}       if $opts{'set-title'};
 		$fh->{songnum}      = 1+$addcount              if $opts{'set-songnum'};
-		if(defined($opts{artwork})) {
+		if($awdb_image_prepared) {
 			$fh->{has_artwork} = 1;
 			$fh->{artworkcnt}  = 1;
 			$fh->{dbid_1}      = $AWDB->InjectImage;
@@ -327,8 +327,21 @@ sub startup {
 	return $fatal_error;
 }
 
-
-
+#############################################################
+# Preapare and add image to artwork database
+sub add_image_to_awdb {
+	my ($filename) = @_;
+	if( $awdb_image_prepared ) {
+		warn "! [****] Skipping $filename because there is already one prepared.\n";
+		return 0;
+	}
+	my $count = $AWDB->PrepareImage(File=>$filename, Model=>$opts{model});
+	if( $count ) {; 
+		$AWDB->LoadArtworkDb or die "Failed to load artwork database\n";
+		$awdb_image_prepared = 1;
+	}
+	return $count;
+}
 #############################################################
 # Add item to playlist
 sub create_playlist_now {
@@ -380,11 +393,69 @@ sub newpl {
 # Calls curl to get files
 sub PODCAST_fetch {
 	my($url,$prefix) = @_;
+	print "* [HTTP] Downloading $url ...\n";
 	my $tmpout = GNUpod::FooBar::get_u_path($prefix,"");
 	my $return = system("curl", "-s", "-L", "-o", $tmpout, $url);
 	return{file=>$tmpout, status=>$return};
 }
 
+sub PODCAST_fetch_media {
+	my($url,$prefix,$length) = @_;
+	if ($opts{'podcast-cache-dir'}) {
+	
+		my @cachefilecandidates = ();
+		my $deepcachefile = $opts{'podcast-cache-dir'}."/".PODCAST_get_sane_path_from_url($url , "");
+		push @cachefilecandidates, $deepcachefile  if $deepcachefile;
+		
+		my $flatcachefile = $opts{'podcast-cache-dir'}."/".PODCAST_strictly_sanitze_path_element((split(/\//, $url))[-1], "cachefile");
+		push @cachefilecandidates, $flatcachefile;
+
+		foreach my $cachefile (@cachefilecandidates) {
+			if ( -e $cachefile && -r $cachefile ) {
+				my $sizedelta = int($length) - int((stat($cachefile))[7]) ;
+				if ( ($length != 0) && (abs($sizedelta) > ($length * 0.05)) ) {
+					print "* [HTTP] Not using cached file $cachefile ... (".abs($sizedelta)." bytes too ".($sizedelta > 0 ? "small" : "big").")\n";
+				} else {
+					print "* [HTTP] Using cached file $cachefile (size:".(stat($cachefile))[7].") ...".
+						(($length != 0 && $sizedelta) ? " (even though it is ".abs($sizedelta)." bytes too " . ($sizedelta>0?"small":"big") . ")" : "")."\n";
+					return {file=>$cachefile, status=>0};
+				}
+			}
+		}
+		print "* [HTTP] Downloading $url ...\n";
+		my $return = system("curl", "-s", "-L", "--create-dirs", "-o" , $deepcachefile, $url);
+		return {file=>$deepcachefile, status=>$return};
+	}
+	else {
+		return PODCAST_fetch($url,$prefix);
+	}
+}
+
+sub PODCAST_strictly_sanitze_path_element {
+	my ($name,$default) = @_;
+	$name =~ s/[^.0-9a-zA-z()_-]/_/g; # limit valid character set
+	$name =~ s/^[.]*//g; #remove leading dots
+	$name =~ s/[.]*$//g; #remove trailing dots (cause problems on windows i heard
+	$name = $default unless $name; #default if empty
+	return $name;
+}
+	
+
+sub PODCAST_get_sane_path_from_url {
+	my($uri,$default) = @_;
+	my($scheme, $authority, $path, $query, $fragment) = $uri =~ m|(?:([^:/?#]+):)?(?://([^/?#]*))?([^?#]*)(?:\?([^#]*))?(?:#(.*))?|;
+	my @pathelements = ($authority, split (/\//, $path));
+	my @cleanpathelements=();
+	foreach my $pe ( @pathelements ) {
+		push @cleanpathelements, PODCAST_strictly_sanitze_path_element($pe,'');
+	}
+	my $cleanpath =  join ("/", @cleanpathelements);
+	$cleanpath =~ s|/[/]+|/|g; # collaps multiple /
+	$cleanpath = $default  if ! $cleanpath;
+	$cleanpath = $default  if $cleanpath eq "/";
+	return $cleanpath;
+}
+
 #############################################################
 #Eventer for START:
 # -> Push array if we found a new item beginning
@@ -393,6 +464,7 @@ sub PODCAST_fetch {
 sub podcastStart {
 	my($hr,$el,@it) = @_;
 	my $hashref_key = $hr->{Base};
+	undef($hr->{cdatabuffer});
 	if($hr->{Context}[-2] eq "rss" &&
 	   $hr->{Context}[-1] eq "channel" &&
 		 $el eq "item") {
@@ -411,14 +483,73 @@ sub podcastStart {
 #Eventer for <foo>CONTENT</foo>
 # => Fillsup %podcast_infos
 sub podcastChar {
+	my($hr,$el) = @_;
+	$hr->{cdatabuffer} .= $el;
+}
+
+#############################################################
+#Eventer for END
+# => Fillsup %podcast_infos
+sub podcastEnd {
 	my($hr,$el) = @_;
 	my $hashref_key = $hr->{Base};
-	if($hr->{Context}[-4] eq "rss" &&
-	   $hr->{Context}[-3] eq "channel" &&
-	   $hr->{Context}[-2] eq "item") {
-		my $ccontext = $hr->{Context}[-1];
-		${$podcast_infos{$hashref_key}}[-1]->{$ccontext}->{"\0"} ||= $el;
+	if(defined($hr->{cdatabuffer}) &&
+	   $hr->{Context}[-3] eq "rss" &&
+	   $hr->{Context}[-2] eq "channel" &&
+	   $hr->{Context}[-1] eq "item") {
+		${$podcast_infos{$hashref_key}}[-1]->{$el}->{"\0"} ||= $hr->{cdatabuffer};
 	}
+	undef($hr->{cdatabuffer});
+}
+
+#############################################################
+#Eventer for START:
+# -> Push array if we found a new item beginning
+# -> Add '<foo bar=barz oink=yak />' stuff to the hash
+# => Fillsup %podcast_channel_infos
+sub podcastChannelStart {
+	my($hr,$el,@it) = @_;
+	my $hashref_key = $hr->{Base};
+	$hr->{cdatabuffer} = undef;
+	if($hr->{Context}[-1] eq "rss" &&
+	   $el eq "channel") {
+		push(@{$podcast_channel_infos{$hashref_key}}, {});
+	}elsif($hr->{Context}[-2] eq "rss" &&
+	   $hr->{Context}[-1] eq "channel" &&
+	   $el ne "item") {
+		if (@it) {
+			my $xref = GNUpod::XMLhelper::mkh($el,@it);
+			${$podcast_channel_infos{$hashref_key}}[-1]->{$el} ||= $xref->{$el};
+		}
+	}
+}
+
+#############################################################
+#Eventer for <foo>CONTENT</foo>
+# => Fillsup %podcast_channel_infos
+sub podcastChannelChar {
+	my($hr,$el) = @_;
+	$hr->{cdatabuffer} .= $el;
+}
+
+#############################################################
+#Eventer for END
+# => Fillsup %podcast_channel_infos
+sub podcastChannelEnd {
+	my($hr,$el) = @_;
+	my $hashref_key = $hr->{Base};
+	if(defined($hr->{cdatabuffer}) &&
+	   $hr->{Context}[-2] eq "rss" &&
+	   $hr->{Context}[-1] eq "channel" &&
+	   $el ne "item") {
+		${$podcast_channel_infos{$hashref_key}}[-1]->{$el}->{"\0"} ||= $hr->{cdatabuffer};
+	}elsif(defined($hr->{cdatabuffer}) &&
+	   $hr->{Context}[-3] eq "rss" &&
+	   $hr->{Context}[-2] eq "channel" &&
+	   $hr->{Context}[-1] ne "item") {
+		${$podcast_channel_infos{$hashref_key}}[-1]->{$hr->{Context}[-1]}->{$el}->{"\0"} ||= $hr->{cdatabuffer};
+	}
+	$hr->{cdatabuffer} = undef; # make sure it doesn't get added to the parent element as well
 }
 
 #############################################################
@@ -448,11 +579,24 @@ sub resolve_podcasts {
 			}
 			#Add the stuff to %podcast_infos and unlink the file after this.
 			eval {
-				my $px = new XML::Parser(Handlers=>{Start=>\&podcastStart, Char=>\&podcastChar});
+				my $px = new XML::Parser(Handlers=>{Start=>\&podcastStart, Char=>\&podcastChar, End=>\&podcastEnd});
 				$px->parsefile($pcrss->{file});
+				my $py = new XML::Parser(Handlers=>{Start=>\&podcastChannelStart, Char=>\&podcastChannelChar, End=>\&podcastChannelEnd});
+				$py->parsefile($pcrss->{file});
 			};
 			warn "! [HTTP] Error while parsing XML: $@\n" if $@;
 			unlink($pcrss->{file}) or warn "Could not unlink $pcrss->{file}, $!\n";
+
+			#Limit the number of podcasts to dowload.
+			my @pods   = @{$podcast_infos{$pcrss->{file}}};
+			my $flimit = int($opts{'podcast-files-limit'});
+			if ($flimit > 0) {
+				splice(@pods, $flimit);
+			} elsif ($flimit < 0) {
+				splice(@pods, 0, $flimit);
+			}				
+			$podcast_infos{$pcrss->{file}} = \@pods;
+
 			$per_file_info{$pcrss->{file}}->{REAL_RSS} = $cf;
 		}
 		else {
@@ -460,11 +604,37 @@ sub resolve_podcasts {
 		}
 	}
 
+#	use Data::Dumper;
+#	print Dumper(\%podcast_channel_infos);
+
 	foreach my $key (keys(%podcast_infos)) {
 		my $cref = $podcast_infos{$key};
+		my $channel = $podcast_channel_infos{$key}[0]; # assuming for now that there's only one channel in the feed
+		# get the artwork
+		if (defined($opts{'podcast-artwork'})) {
+			my $channel_image_url = ( $channel->{"itunes:image"}->{"href"} or
+				$channel->{"image"}->{"url"}->{"\0"} ) ;
+			if ( $channel_image_url ) {
+				my $channel_image = PODCAST_fetch_media($channel_image_url, "/tmp/gnupodcast_image", 0);
+				if($channel_image->{status} or (!(-f $channel_image->{file}))) {
+					warn "! [HTTP] Failed to download $channel_image to ".$channel_image->{file}."\n";
+				}
+				else {
+					add_image_to_awdb($channel_image->{file});
+					if ( ! $opts{'podcast-cache-dir'} ) {
+						unlink($channel_image->{file}) or warn "Could not unlink ".$channel_image->{file}.", $!\n";
+					}
+				}
+			}
+		}
+
 		foreach my $podcast_item (@$cref) {
 			my $c_title = $podcast_item->{title}->{"\0"};
-			my $c_author = $podcast_item->{author}->{"\0"};
+			my $c_author = ( $podcast_item->{"itunes:author"}->{"\0"} or
+					 $podcast_item->{author}->{"\0"} or
+					 $channel->{"itunes:author"}->{"\0"} or
+					 $channel->{"managingEditor"}->{"\0"} );
+			my $c_rdate = $podcast_item->{pubDate}->{"\0"}; 
 			my $c_desc  = $podcast_item->{description}->{"\0"};
 			my $c_url   = $podcast_item->{enclosure}->{url};
 			#We use the URL as GUID if there isn't one...			
@@ -480,14 +650,13 @@ sub resolve_podcasts {
 				warn "! [HTTP] Podcast $c_url ($c_title) exists, no need to download this file\n";
 				next;
 			}		
-			print "* [HTTP] Downloading $c_url ...\n";
-			my $rssmedia = PODCAST_fetch($c_url, "/tmp/gnupodcast_media");
+			my $rssmedia = PODCAST_fetch_media($c_url, "/tmp/gnupodcast_media", $podcast_item->{enclosure}->{length});
 			if($rssmedia->{status} or (!(-f $rssmedia->{file}))) {
 				warn "! [HTTP] Failed to download $c_url to $rssmedia->{file}\n";
 				next;
 			}
 			
-			$per_file_info{$rssmedia->{file}}->{UNLINK}    = 1;  # Remote tempfile
+			$per_file_info{$rssmedia->{file}}->{UNLINK}    = 1 unless $opts{'podcast-cache-dir'};  # Remove tempfile if not caching
 			$per_file_info{$rssmedia->{file}}->{ISPODCAST} = 1;  # Triggers mediatype fix
 			
 			# Set information/tags from XML-File
@@ -496,6 +665,7 @@ sub resolve_podcasts {
 			$per_file_info{$rssmedia->{file}}->{title}       = $c_title   if $c_title;
 			$per_file_info{$rssmedia->{file}}->{artist}      = $c_author  if $c_author;
 			$per_file_info{$rssmedia->{file}}->{desc}        = $c_desc    if $c_desc;
+			$per_file_info{$rssmedia->{file}}->{releasedate} = int(Date::Parse::str2time($c_rdate))+MACTIME    if $c_rdate;
 			
 			push(@files,$rssmedia->{file});
 		}
@@ -547,6 +717,10 @@ Usage: gnupod_addsong.pl [-h] [-m directory] File1 File2 ...
    -d, --duplicate                  Allow duplicate files
    -p, --playlist=string            Add songs to this playlist, can be used multiple times
        --playlist-is-podcast        Set podcast flag for playlist(s) created using '--playlist'
+       --podcast-artwork            Download and install artwork for podcasts from their channel.
+       --podcast-cache-dir=string   Set a directory in which podcast media files will be cached.
+       --podcast-files-limit=int    Limit the number of files that are downloaded.
+                                    0 = download all (default), -X = download X oldest items, X = download X newest items
        --disable-v1                 Do not read ID3v1 Tags (MP3 Only)
        --disable-v2                 Do not read ID3v2 Tags (MP3 Only)
    -x  --decode=pcm|mp3|aac|aacbm   Convert FLAC Files to WAVE/MP3 or AAC 'on-the-fly'. Use '-e' to specify a quality/bitrate
@@ -578,8 +752,8 @@ EOF
 
 sub version {
 die << "EOF";
-gnupod_addsong.pl (gnupod) 0.99.6
-Copyright (C) Adrian Ulrich 2002-2007
+gnupod_addsong.pl (gnupod) 0.99.7
+Copyright (C) Adrian Ulrich 2002-2008
 
 This is free software; see the source for copying conditions.  There is NO
 warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
